#ifndef SCRIPT_H
#define SCRIPT_H TRUE

struct Script {
   OBJECT_HEADER
   OBJECTID TargetID;             /* The object that script objects must be initialised to */
   OBJECTID PrimaryObjectID;      /* Container that post-child objects initialised to the script belong to */
   OBJECTID OutputID;             /* Output object for sending user messages */
   LONG     Static;               /* Set to TRUE if the Script is static */
   LONG     TotalArgs;            /* Total number of arguments in the LocalArgs table */
   LONG     TotalObjects;         /* Total number of objects listed in the ObjectList array */
   LONG     Repeat;               /* Set this field to execute the script multiple times */
   LONG     RepeatStart;          /* Optional setting to start the repeat count from a non-zero number */
   LONG     Flags;                /* Optional flags */
   ERROR    Error;                /* If an error occurred, this field will indicate the error number */
   LARGE    FileSize;             /* The byte-size of the original script file */
   LARGE    TimeStamp;            /* The timestamp of the original script file */

   /*** Unpublished fields ***/
   STRING   String;               /* String to translate if no file location */
   STRING   Location;             /* File location of the script */
   STRING   Procedure;            /* Procedure to execute */
   STRING   Service;              /* The service that you want the script to execute under */
   LONG     GeneratedID;          /* The unique ID generated for {id} replacement */
   MEMORYID LocationMID;
   MEMORYID StringMID;
   MEMORYID LocalArgsMID;
   MEMORYID ObjectListMID;        /* List of objects that were initialised by the script (OBJECTID *) */
   MEMORYID ProcedureMID;
   MEMORYID *LocalArgs;           /* Argument table (created via calls to SetUnlistedField()) */
   OBJECTID *ObjectList;
   MEMORYID ServiceMID;
   UBYTE    Language[4];          /* 3-character language code, null-terminated */
   UBYTE    LanguageDir[32];      /* Directory to use for language files */

   /*** Private base-class fields ***/
   struct XML *XML;
   LONG ActivationCount;  /* Remember the total number of times the script has been activated */
   LONG LoopIndex;        /* Internal loop index for <repeat> statements */
   LONG LoopEnd;          /* Internal loop end for <repeat> statements */
   ERROR (*ChildFeedback)(OBJECTID, struct XMLTag *, STRING, STRING);
   WORD prvUnique;
   UBYTE prvQuit;
   UBYTE prvTranslateAll;
   OBJECTID ScriptOwnerID;
};

#endif /* SCRIPT_H */
