/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      FontServer
Version:   1.0
ID:        ID_FONTSERVER
Include:   graphics/font.h
Category:  Graphics
Date:      December 2002
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2002. All rights reserved.
Short:     Manages the fonts that are installed in the system.
Keywords:  font, graphics, server, manage, list, query

###DESCRIPTION###

###END###

*****************************************************************************/

static void ScanTrueTypeDirectory(OBJECTPTR);
static void ScanFixedDirectory(OBJECTPTR);
static ERROR AnalyseBitmapFont(STRING, struct winfnt_header_fields *, STRING *);

struct FontEntry {
   UBYTE Name[40];
   LONG FixedPoints[20];
   UBYTE Styles[10][30];
   LONG Flags;
};

static ERROR GET_TotalFonts(struct FontServer *, LONG *);

static struct FieldArray FontServerFields[] = {

   /*** Virtual fields ***/
   { "TotalFonts", 0, FDF_LONG|FDF_R, 0, GET_TotalFonts, NULL },
   END_FIELD
};

/*****************************************************************************
** Actions
*/

static ERROR FONTSERVER_Free(struct FontServer *, APTR);
static ERROR FONTSERVER_GetUnlistedField(struct FontServer *, struct acGetUnlistedField *);
static ERROR FONTSERVER_Init(struct FontServer *, APTR);
static ERROR FONTSERVER_NewObject(struct FontServer *, APTR);
static ERROR FONTSERVER_Refresh(struct FontServer *, APTR);

static struct ActionArray FontServerActions[] = {
   { AC_Free,             FONTSERVER_Free },
   { AC_GetUnlistedField, FONTSERVER_GetUnlistedField },
   { AC_Init,             FONTSERVER_Init },
   { AC_NewObject,        FONTSERVER_NewObject },
   { AC_Refresh,          FONTSERVER_Refresh },
   { NULL, NULL }
};

/*****************************************************************************
** Methods
*/

static ERROR FONTSERVER_InstallFont(struct FontServer *, struct mtInstallFont *);
static ERROR FONTSERVER_RemoveFont(struct FontServer *, struct mtRemoveFont *);
static ERROR FONTSERVER_SelectFont(struct FontServer *, struct mtSelectFont *);

static struct FunctionField argsInstallFont[] = { { "Files", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsRemoveFont[]  = { { "Name", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsSelectFont[]  = { { "Name", ARG_STR }, { "Style", ARG_STR }, { "Point", ARG_LONG }, { "Flags", ARG_LONG }, { "Location", ARG_STR|ARG_RESULT }, { "Handle", ARG_LONG|ARG_RESULT }, { NULL, NULL } };

static struct MethodArray FontServerMethods[] = {
   { MT_InstallFont,   FONTSERVER_InstallFont, "InstallFont", argsInstallFont, sizeof(struct mtInstallFont) },
   { MT_RemoveFont,    FONTSERVER_RemoveFont,  "RemoveFont",  argsRemoveFont,  sizeof(struct mtRemoveFont) },
   { MT_SelectFont,    FONTSERVER_SelectFont,  "SelectFont",  argsSelectFont,  sizeof(struct mtSelectFont) },
   { NULL, NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Function: AddFontClass()
*/

static ERROR AddFontServerClass(void)
{
   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FontServerClass, NULL,
      FID_BaseClassID|TLONG, ID_FONTSERVER,
      FID_Version|TFLOAT,    VER_FONTSERVER,
      FID_Name|TSTRING,      "FontServer",
      FID_Category|TLONG,    CCF_GRAPHICS,
      FID_Actions|TPTR,      FontServerActions,
      FID_Methods|TPTR,      FontServerMethods,
      FID_Fields|TPTR,       FontServerFields,
      FID_Size|TLONG,        sizeof(struct FontServer),
      TAGEND));
}

/*****************************************************************************
** FontServer: Free
*/

static ERROR FONTSERVER_Free(struct FontServer *Self, APTR Void)
{
   if (Self->prvConfigID) {
      ActionMsg(AC_Free, Self->prvConfigID, NULL);
      Self->prvConfigID = NULL;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** FontServer: GetUnlistedField
*/

static ERROR FONTSERVER_GetUnlistedField(struct FontServer *Self, struct acGetUnlistedField *Args)
{
   struct mtGetConfigSectionFromIndex getsection;
   struct mtReadConfig read;
   OBJECTPTR config;
   LONG index, i, pos, stylecount;

   if (AccessObject(Self->prvConfigID, 3000, &config) IS ERR_Okay) {
      if (StrCompare("FontName(", Args->Field, 0, NULL) IS ERR_Okay) {
         getsection.Index = StrToInt(Args->Field);
         if (Action(MT_GetConfigSectionFromIndex, config, &getsection) IS ERR_Okay) {
            read.Section = getsection.Result;
            read.Item = "Name";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               StrCopy(read.Data, Args->Buffer, Args->Size);
            }
         }
         else {
            ReleaseObject(config);
            return(ObjectError(Self, ERH_GetUnlistedField, ERR_OutOfRange));
         }
      }
      else if (StrCompare("FontStyle(", Args->Field, 0, NULL) IS ERR_Okay) {
         getsection.Index = StrToInt(Args->Field);
         if (Action(MT_GetConfigSectionFromIndex, config, &getsection) IS ERR_Okay) {
            read.Section = getsection.Result;
            read.Item = "Styles";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               for (i=10; Args->Field[i]; i++) if ((Args->Field[i] >= '0') AND (Args->Field[i] <= '9')) break;
               while ((Args->Field[i] >= '0') AND (Args->Field[i] <= '9')) i++;
               index = StrToInt(Args->Field+i);

               for (i=0; (read.Data[i]) AND (index > 0); i++) {
                  if (read.Data[i] IS ',') index--;
               }

               pos = 0;
               while ((read.Data[i]) AND (read.Data[i] != ',') AND (pos < Args->Size-1)) Args->Buffer[pos++] = read.Data[i++];
               Args->Buffer[pos] = 0;
            }
         }
         else {
            ReleaseObject(config);
            return(ObjectError(Self, ERH_GetUnlistedField, ERR_OutOfRange));
         }
      }
      else if (StrCompare("StyleCount(", Args->Field, 0, NULL) IS ERR_Okay) {
         getsection.Index = StrToInt(Args->Field);
         if (Action(MT_GetConfigSectionFromIndex, config, &getsection) IS ERR_Okay) {
            read.Section = getsection.Result;
            read.Item = "Styles";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               stylecount = 1;
               for (i=0; read.Data[i]; i++) if (read.Data[i] IS ',') stylecount++;
               IntToStr(stylecount, Args->Buffer, Args->Size);
            }
            else IntToStr(0, Args->Buffer, Args->Size);
         }
         else {
            ReleaseObject(config);
            return(ObjectError(Self, ERH_GetUnlistedField, ERR_OutOfRange));
         }
      }
      else {
         DPrintF("@GetUnlistedField:","Unrecognised field name \"%s\".", Args->Field);
         ReleaseObject(config);
         return(ERR_Failed);
      }

      ReleaseObject(config);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetUnlistedField, ERR_AccessObject));
}

/*****************************************************************************
** FontServer: Init
*/

static ERROR FONTSERVER_Init(struct FontServer *Self, APTR Void)
{
   OBJECTPTR config;
   LONG type;
   UBYTE refresh;

   /*** Create the font registry file if it does not exist ***/

   if ((AnalyseLocation("fonts:registry.cfg", &type) != ERR_Okay) OR (type != LOC_FILE)) {
      refresh = TRUE;
   }
   else refresh = FALSE;

   /*** Load the fonts:registry.cfg file ***/

   if (CreateObject(ID_CONFIG, Self->Head.ObjectFlags|NF_CHILD, &config, &Self->prvConfigID,
      FID_Location|TSTRING, "fonts:registry.cfg",
      TAGEND) IS ERR_Okay) {

      if (refresh IS TRUE) Action(AC_Refresh, Self, NULL);

      ReleaseObject(config);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Init, ERR_CreateObject));
}
/*****************************************************************************

###METHOD###
Name:  InstallFont()
Short: Installs a new font to the system.

###DESCRIPTION###
<p>The InstallFont() function is used to install new fonts on a system
running the Pandora Engine.  While it is possible for users to download new
font files and install them by hand, this is a process that is too difficult
for novices and is open to mistakes on the part of the user.  By using a
program that uses the InstallFont() function, the installation process can
take place automatically.</p>

<p>To install a new font, you only need to know the location of the font
file(s).  The rest of the information about the font will be derived after
an analysis of the data.</p>

<p>Once this function is called, the data files will be copied into the
correct sub-directory and the font registration files will be updated to
reflect the presence of the new font.  The font will be available immediately
thereafter, so there is no need to reset the system to acknowledge the
presence of the font.</p>

###INPUT###
STRING/Files: A list of the font files that are to be installed must be specified here.  If there is more than one data file, separate each file name with a semi-colon.

###RESULT###
ERR_Okay: The font information was successfully installed.
ERR_ExclusiveDenied: Access to the SystemFonts object was denied.
ERR_Args: Invalid arguments were detected.
ERR_NoSupport: One of the font files is in an unsupported file format.

###SEE ALSO###
Method: RemoveFont()

###END###

*****************************************************************************/

static ERROR FONTSERVER_InstallFont(struct FontServer *Self, struct mtInstallFont *Args)
{
   struct mtCopyFile copy;
   struct acRead read;
   BYTE buffer[512];
   OBJECTPTR file;
   LONG i, j;
   STRING directory;

   /*** Check that we received the minimum required arguments ***/

   if ((!Args) OR (!Args->Files)) return(ObjectError(Self, MT_InstallFont, ERR_Args));

   DPrintF("~InstallFont()","[FontServer:%d] Files: %s", Self->Head.UniqueID, Args->Files);

   /*** Copy all files to the destination directory ***/

   i = 0;
   while (Args->Files[i]) {
      for (j=0; (Args->Files[i]) AND (Args->Files[i] != ';'); j++) buffer[j] = Args->Files[i++];
      buffer[j++] = 0;

      /* Read the file header to figure out whether the file belongs in the
      ** fixed or truetype directory.
      */

      if (CreateObject(ID_FILE, NULL, &file, NULL,
         FID_Flags|TLONG,      FL_READ,
         FID_Location|TSTRING, buffer,
         TAGEND) IS ERR_Okay) {

         directory = "fixed";
         read.Buffer = buffer;
         read.Length = 256;
         if (Action(AC_Read, file, &read) IS ERR_Okay) {
            if ((buffer[0] IS 'M') AND (buffer[1] IS 'Z')) directory = "fixed";
            else directory = "truetype";

            StrFormat(buffer, sizeof(buffer), "fonts:%s/", directory);
            copy.Destination = buffer;
            Action(MT_CopyFile, file, &copy);
         }

         Action(AC_Free, file, NULL);
      }

      if (Args->Files[i]) {
         i++;
         while ((Args->Files[i]) AND (Args->Files[i] <= 0x20)) i++;
      }
   }

   /*** Refresh the font server so that the installed files will show up in the font list ***/

   Action(AC_Refresh, Self, NULL);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** FontServer: NewObject
*/

static ERROR FONTSERVER_NewObject(struct FontServer *Self, APTR Void)
{
   SetName(Self, "SystemFonts");
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Refreshes the system font list with up-to-date font information.

###DESCRIPTION###
<p>The RefreshFonts() function is used to scan for fonts that have been
installed in the system, for the purpose of detecting deleted and recently
installed font files.  Ideally this function should not be necessary when
the InstallFont() and RemoveFont() methods are used correctly, however
it can be useful when font files have been manually deleted or added to the
system.</p>

<p>Refreshing fonts can take an extensive amount of time as each font file
needs to be completely analysed for information.  Once the analysis is
complete, the "SystemFonts" object will be updated and the "fonts:registry.cfg"
file will reflect current font settings.</p>

###RESULT###
ERR_Okay: Fonts were successfully refreshed.
ERR_AccessObject: Access to the SytsemFonts object was denied, or the object does not exist.

###END###

*****************************************************************************/

static ERROR FONTSERVER_Refresh(struct FontServer *Self, APTR Void)
{
   #define MAX_STYLES 20
   struct mtWriteConfig add;
   struct ConfigEntry *entries;
   OBJECTPTR config, file;
   STRING section, styles[MAX_STYLES];
   LONG i, pos, stylecount, amtentries, j;

   DPrintF("~Refresh()","[FontServer:%d] Refreshing the fonts: directory.", Self->Head.UniqueID);

   if (AccessObject(Self->prvConfigID, 3000, &config) != ERR_Okay) {
      ObjectError(Self, ERH_Refresh, ERR_AccessObject);
      StepBack();
      return(ERR_AccessObject);
   }

   Action(AC_Clear, config, NULL); /* Clear out existing font information */

   ScanFixedDirectory(config);

   ScanTrueTypeDirectory(config);

   /*** Sort the font names into alphabetical order ***/

   Action(MT_SortConfig, config, NULL);

   /*** Create a style list for each font ***/

   DPrintF("Refresh:","Generating style lists for each font.");

   if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                         FID_Entries|TPTR,     &entries,
                         TAGEND) IS ERR_Okay) {

      section = entries[0].Section;
      stylecount = 0;
      i = 0;
      while (i < amtentries) {
         if ((i < amtentries-1) AND (StrCompare(entries[i].Section, section, 0, STR_MATCHLENGTH|STR_CASE) IS ERR_Okay)) {
            /*** If this is a style item, add it to our style list ***/
            if (StrCompare("Fixed:", entries[i].Item, 6, NULL) IS ERR_Okay) {
               if (stylecount < MAX_STYLES-1) styles[stylecount++] = entries[i].Item + 6;
            }
            else if (StrCompare("Scale:", entries[i].Item, 6, NULL) IS ERR_Okay) {
               if (stylecount < MAX_STYLES-1) styles[stylecount++] = entries[i].Item + 6;
            }
            i++;
         }
         else {
            /*** Write the style list to the font configuration ***/
            UBYTE buffer[300];
            UBYTE sectionstr[80];

            if (stylecount > 0) {
               styles[stylecount] = NULL;
               StrSort(styles, SBF_NODUPLICATES);

               pos = 0;
               for (j=0; styles[j]; j++) {
                  if ((pos > 0) AND (pos < sizeof(buffer)-1)) buffer[pos++] = ',';
                  pos += StrCopy(styles[j], buffer+pos, sizeof(buffer)-pos);
               }

               StrCopy(section, sectionstr, sizeof(sectionstr));

               add.Section = sectionstr;
               add.Item = "Styles";
               add.Data = buffer;
               Action(MT_WriteConfig, config, &add);

               /*** Reset the config index since we added a new entry to the object ***/

               if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                                     FID_Entries|TPTR,     &entries,
                                     TAGEND) != ERR_Okay) break;

               for (i=0; i < amtentries; i++) {
                  if (StrCompare(entries[i].Section, sectionstr, 0, STR_MATCHLENGTH|STR_CASE) IS ERR_Okay) {
                     while ((i < amtentries) AND (StrCompare(entries[i].Section, sectionstr, 0, STR_MATCHLENGTH|STR_CASE) IS ERR_Okay)) i++;
                     if (i < amtentries) section = entries[i].Section;
                     break;
                  }
               }
            }
            else {
               section = entries[i].Section;
               i++;
            }
            stylecount = 0;
         }
      }
   }

   /*** Save the font configuration file ***/

   DPrintF("Refresh:","Saving the font configuration file.");

   if (CreateObject(ID_FILE, NULL, &file, NULL,
      FID_Location|TSTRING, "fonts:registry.cfg",
      FID_Flags|TLONG,      FL_NEW|FL_WRITE,
      TAGEND) IS ERR_Okay) {
      ActionTags(AC_SaveToObject, config, file->UniqueID);
      Action(AC_Free, file, NULL);
   }

   ReleaseObject(config);
   StepBack();
   return(ERR_Okay);
}


/*****************************************************************************
** Internal: ScanTrueTypeDirectory()
*/

static void ScanTrueTypeDirectory(OBJECTPTR Config)
{
   struct FileInfo *scan, *list;
   LONG j, n, style;
   UBYTE location[100];
   UBYTE buffer[200];
   UBYTE section[200];
   FT_Face ftface;
   struct mtWriteConfig add;
   FT_Open_Args open;

   DPrintF("~FontServer:","Scanning for truetype fonts.");

   if (ReadDirectory("fonts:truetype/", RDF_FILE, &list) IS ERR_Okay) {
      for (scan=list; scan; scan=scan->Next) {
         StrFormat(location, sizeof(location), "fonts:truetype/%s", scan->Name);

         for (j=0; location[j]; j++);
         while ((j > 0) AND (location[j-1] != '.') AND (location[j-1] != ':') AND (location[j-1] != '/') AND (location[j-1] != '\\')) j--;

         ResolveAssign(location, (STRING *)&open.pathname, NULL);
         open.flags    = FT_OPEN_PATHNAME;
         if (!FT_Open_Face(glFTLibrary, &open, 0, &ftface)) {
            FreeMemory(open.pathname);

            DPrintF("FontServer:","Detected font file \"%s\", name: %s, style: %s", location, ftface->family_name, ftface->style_name);

            if (ftface->family_name) {
               n = StrCopy(ftface->family_name, section, sizeof(section));
            }
            else {
               n = 0;
               while ((j > 0) AND (location[j-1] != ':') AND (location[j-1] != '/') AND (location[j-1] != '\\')) j--;
               while ((location[j]) AND (location[j] != '.')) section[n++] = location[j++];
            }
            section[n] = 0;

            /*** Strip any style references out of the font name and keep them as style flags ***/

            style = NULL;
            if (ftface->style_name) {
               if ((n = StrSearch(" Bold", section, STR_MATCHCASE)) != -1) {
                  for (j=0; " Bold"[j]; j++) section[n++] = ' ';
                  style |= FTF_BOLD;
               }

               if ((n = StrSearch(" Italic", section, STR_MATCHCASE)) != -1) {
                  for (j=0; " Italic"[j]; j++) section[n++] = ' ';
                  style |= FTF_ITALIC;
               }
            }

            for (n=0; section[n]; n++);
            while ((n > 0) AND (section[n-1] <= 0x20)) n--;
            section[n] = 0;

            add.Section = section;
            add.Item    = "Name";
            add.Data    = section;
            Action(MT_WriteConfig, Config, &add);

            if (FT_IS_SCALABLE(ftface)) {
               add.Item    = "Scalable";
               add.Data    = "Yes";
               Action(MT_WriteConfig, Config, &add);
            }

            /*** Add the style with a link to the font file location ***/

            if (FT_IS_SCALABLE(ftface)) {
               if ((ftface->style_name) AND (StrCompare("regular", ftface->style_name, 0, STR_MATCHLENGTH) != ERR_Okay)) {
                  CharCopy("Scale:", buffer, sizeof(buffer));
                  StrCopy(ftface->style_name, buffer+6, sizeof(buffer)-6);
                  add.Item    = buffer;
                  add.Data    = location;
                  Action(MT_WriteConfig, Config, &add);
               }
               else {
                  if (style IS FTF_BOLD) add.Item = "Scale:Bold";
                  else if (style IS FTF_ITALIC) add.Item = "Scale:Italic";
                  else if (style IS (FTF_BOLD|FTF_ITALIC)) add.Item = "Scale:Bold Italic";
                  else add.Item = "Scale:Regular";

                  add.Data    = location;
                  Action(MT_WriteConfig, Config, &add);
               }
            }

            FT_Done_Face(ftface);
         }
         else {
            FreeMemory(open.pathname);
            DPrintF("@FontServer:","Failed to analyse font file \"%s\".", location);
         }
      }

      FreeMemory(list);
   }
   else DPrintF("@FontServer:","Failed to open the fonts:truetype/ directory.");

   StepBack();
}

/*****************************************************************************
** Internal: ScanFixedDirectory()
*/

static void ScanFixedDirectory(OBJECTPTR Config)
{
   struct FileInfo *list, *scan;
   struct mtWriteConfig add;
   struct winfnt_header_fields header;
   LONG j, n, style;
   UBYTE location[100];
   UBYTE section[200];
   STRING facename;
   UBYTE bold, bolditalic, italic;

   DPrintF("~FontServer:","Scanning for fixed fonts.");

   bold = FALSE;
   bolditalic = FALSE;
   italic = FALSE;

   if (ReadDirectory("fonts:fixed/", RDF_FILE, &list) IS ERR_Okay) {
      for (scan=list; scan; scan=scan->Next) {
         StrFormat(location, sizeof(location), "fonts:fixed/%s", scan->Name);

         if (AnalyseBitmapFont(location, &header, &facename) IS ERR_Okay) {

            DPrintF("6Font:","Detected font file \"%s\", name: %s", location, facename);

            if (!facename) continue;
            StrCopy(facename, section, sizeof(section));

            /*** Strip any style references out of the font name and keep them as style flags ***/

            style = NULL;
            if ((n = StrSearch(" Bold", section, STR_MATCHCASE)) != -1) {
               for (j=0; " Bold"[j]; j++) section[n++] = ' ';
               style |= FTF_BOLD;
            }

            if ((n = StrSearch(" Italic", section, STR_MATCHCASE)) != -1) {
               for (j=0; " Italic"[j]; j++) section[n++] = ' ';
               style |= FTF_ITALIC;
            }

            if (header.italic) style |= FTF_ITALIC;
            if (header.weight >= 600) style |= FTF_BOLD;

            for (n=0; section[n]; n++);
            while ((n > 0) AND (section[n-1] <= 0x20)) n--;
            section[n] = 0;

            add.Section = section;
            add.Item    = "Name";
            add.Data    = section;
            Action(MT_WriteConfig, Config, &add);

            /*** Add the style with a link to the font file location ***/

            if (style IS FTF_BOLD) {
               add.Item = "Fixed:Bold";
               add.Data    = location;
               Action(MT_WriteConfig, Config, &add);
               bold = TRUE;
            }
            else if (style IS FTF_ITALIC) {
               add.Item = "Fixed:Italic";
               add.Data    = location;
               Action(MT_WriteConfig, Config, &add);
               bolditalic = TRUE;
            }
            else if (style IS (FTF_BOLD|FTF_ITALIC)) {
               add.Item = "Fixed:Bold Italic";
               add.Data    = location;
               Action(MT_WriteConfig, Config, &add);
               italic = TRUE;
            }
            else {
               add.Item = "Fixed:Regular";
               add.Data    = location;
               Action(MT_WriteConfig, Config, &add);

               if (bold IS FALSE) {
                  add.Item = "Fixed:Bold";
                  Action(MT_WriteConfig, Config, &add);
               }

               if (bolditalic IS FALSE) {
                  add.Item = "Fixed:Bold Italic";
                  Action(MT_WriteConfig, Config, &add);
               }

               if (italic IS FALSE) {
                  add.Item = "Fixed:Italic";
                  Action(MT_WriteConfig, Config, &add);
               }
            }

            FreeMemory(facename);
         }
      }
      FreeMemory(list);
   }

   StepBack();
}

/****************************************************************************/

static ERROR AnalyseBitmapFont(STRING Location, struct winfnt_header_fields *Header, STRING *FaceName)
{
   struct winmz_header_fields mz_header;
   struct winne_header_fields ne_header;
   struct acRead read;
   OBJECTPTR file;
   LONG i, res_offset, font_offset;
   UWORD size_shift, font_count, type_id, count;

   if ((!Location) OR (!Header) OR (!FaceName)) return(ERR_Args);

   *FaceName = NULL;
   if (CreateObject(ID_FILE, NULL, &file, NULL,
      FID_Location|TSTRING, Location,
      FID_Flags|TLONG,      FL_READ,
      TAGEND) IS ERR_Okay) {

      read.Buffer = (APTR)&mz_header;
      read.Length = sizeof(mz_header);
      Action(AC_Read, file, &read);

      if (mz_header.magic IS ID_WINMZ) {
         ActionTags(AC_Seek, file, (DOUBLE)mz_header.lfanew, POS_START);

         read.Buffer = (APTR)&ne_header;
         read.Length = sizeof(ne_header);
         if ((Action(AC_Read, file, &read) IS ERR_Okay) AND (ne_header.magic IS ID_WINNE)) {

            res_offset = mz_header.lfanew + ne_header.resource_tab_offset;
            ActionTags(AC_Seek, file, (DOUBLE)res_offset, POS_START);

            font_count  = 0;
            font_offset = 0;
            size_shift  = ReadWordLE(file);

            for (type_id=ReadWordLE(file); type_id; type_id=ReadWordLE(file)) {
               count = ReadWordLE(file);

               if (type_id IS 0x8008) {
                  font_count  = count;
                  GetField(file, FID_Position, FT_LONG, &font_offset);
                  font_offset = font_offset + 4;
                  break;
               }

               ActionTags(AC_Seek, file, (DOUBLE)(4 + count * 12), POS_CURRENT);
            }

            if ((!font_count) OR (!font_offset)) {
               DPrintF("@FontServer:","There are no fonts in file \"%s\"", Location);
               Action(AC_Free, file, NULL);
               return(ERR_Failed);
            }

            ActionTags(AC_Seek, file, (DOUBLE)font_offset, POS_START);

            {
               struct FontList fonts[font_count];

               for (i=0; i < font_count; i++) {
                  fonts[i].Offset = ReadWordLE(file)<<size_shift;
                  fonts[i].Size   = ReadWordLE(file)<<size_shift;
                  ActionTags(AC_Seek, file, 8.0, POS_CURRENT);
               }

               ActionTags(AC_Seek, file, (DOUBLE)fonts[0].Offset, POS_START);

               /*** Read the font header ***/

               read.Buffer = (APTR)Header;
               read.Length = sizeof(struct winfnt_header_fields);
               if (Action(AC_Read, file, &read) != ERR_Okay) {
                  Action(AC_Free, file, NULL);
                  return(ERR_Read);
               }

               if ((Header->version != 0x200) AND (Header->version != 0x300)) {
                  DPrintF("@FontServer:","Font \"%s\" is written in unsupported version %d.", Location, Header->version);
                  Action(AC_Free, file, NULL);
                  return(ERR_NoSupport);
               }

               if (Header->file_type & 1) {
                  DPrintF("@FontServer:","Font \"%s\" is in the non-supported vector font format.", Location);
                  Action(AC_Free, file, NULL);
                  return(ERR_NoSupport);
               }

               /*** Extract the name of the font ***/

               {
                  #define MAX_FONTFACE 50
                  UBYTE face[MAX_FONTFACE];
                  ActionTags(AC_Seek, file, (DOUBLE)fonts[0].Offset + Header->face_name_offset, POS_START);

                  for (i=0; i < MAX_FONTFACE-1; i++) {
                     read.Buffer = face + i;
                     read.Length = 1;
                     if ((Action(AC_Read, file, &read) != ERR_Okay) OR (!face[i])) {
                        break;
                     }
                  }
                  face[i] = 0;
                  *FaceName = StrClone(face);
               }
            }

            Action(AC_Free, file, NULL);
            return(ERR_Okay);
         }

         Action(AC_Free, file, NULL);
         return(ERR_NoSupport);
      }

      Action(AC_Free, file, NULL); /* File is not a windows fixed font */
      return(ERR_NoSupport);
   }
   else return(ERR_File);
}

/*****************************************************************************

###METHOD###
Name:  RemoveFont()
Short: Removes an installed font from the system.

###DESCRIPTION###
<p>This function is designed to remove fonts that have been installed on a
Pandora Engine system.  All you need to do is provide the Name (face) of the
font and it will be removed if it exists.  Once a font has been removed, the
data is permanently destroyed and it cannot be recovered.</p>

###INPUT###
STRING/Name: The name of the font face that you want to delete.

###RESULT###
ERR_Okay: The font was succesfully removed.
ERR_Args: Invalid arguments were specified.

###END###

*****************************************************************************/

static ERROR FONTSERVER_RemoveFont(struct FontServer *Self, struct mtRemoveFont *Args)
{
   struct ConfigEntry *entries;
   struct mtReadConfig read;
   OBJECTPTR file, config;
   STRING string;
   LONG i, n, amtentries;
   UBYTE buffer[200], style[200];

   if ((!Args) OR (!Args->Name) OR (!Args->Name[0])) return(ObjectError(Self, MT_RemoveFont, ERR_Args));

   DPrintF("~RemoveFont()","%s", Args->Name);

   if (AccessObject(Self->prvConfigID, 3000, &config) != ERR_Okay) {
      ObjectError(Self, MT_RemoveFont, ERR_AccessObject);
      StepBack();
      return(ERR_AccessObject);
   }

   /*** Delete all files related to this font ***/

   if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                         FID_Entries|TPTR,     &entries,
                         TAGEND) != ERR_Okay) {
      ReleaseObject(config);
      ObjectError(Self, MT_RemoveFont, ERR_NoData);
      StepBack();
      return(ERR_NoData);
   }

   for (i=0; i < amtentries; i++) {
      if (StrCompare("Name", entries[i].Item, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         if (StrCompare(Args->Name, entries[i].Data, 0, STR_MATCHLENGTH) IS ERR_Okay) {
            break;
         }
      }
   }

   if (i >= amtentries) {
      ReleaseObject(config);
      ObjectError(Self, MT_RemoveFont, ERR_Search);
      StepBack();
      return(ERR_Search);
   }

   read.Section = entries[i].Section;
   read.Item    = "Styles";
   if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
      string = read.Data;

      DPrintF("RemoveFont:","Scanning styles: %s", string);

      while (*string) {
         for (n=0; ((*string) AND (*string != ',')); n++) style[n] = *string++;
         style[n] = 0;

         if (*string IS ',') string++;

         StrFormat(buffer, sizeof(buffer), "Fixed:%s", style);
         read.Item = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            if (CreateObject(ID_FILE, NULL, &file, NULL,
               FID_Location|TSTRING, read.Data,
               FID_Flags|TLONG,      FL_READ,
               TAGEND) IS ERR_Okay) {

               Action(MT_DeleteFile, file, NULL);

               Action(AC_Free, file, NULL);
            }
            else DPrintF("@RemoveFont:","Failed to open file \"%s\".", read.Data);
         }

         StrFormat(buffer, sizeof(buffer), "Scale:%s", style);
         read.Item = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            if (CreateObject(ID_FILE, NULL, &file, NULL,
               FID_Location|TSTRING, read.Data,
               FID_Flags|TLONG,      FL_READ,
               TAGEND) IS ERR_Okay) {

               Action(MT_DeleteFile, file, NULL);

               Action(AC_Free, file, NULL);
            }
            else DPrintF("@RemoveFont:","Failed to open file \"%s\".", read.Data);
         }
      }
   }
   else DPrintF("@RemoveFont:","There is no Styles entry for the %s font.", Args->Name);

   StrCopy(entries[i].Section, buffer, sizeof(buffer));
   ActionTags(MT_DeleteConfigSection, config, buffer);

   ReleaseObject(config);

   DPrintF("RemoveFont:","Font removed successfully.");
   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  SelectFont
Short: Searches for a 'best fitting' font file based on select criteria.
###END###

*****************************************************************************/

static ERROR FONTSERVER_SelectFont(struct FontServer *Self, struct mtSelectFont *Args)
{
   OBJECTPTR config;
   struct mtReadConfig read;
   struct ConfigEntry *entries;
   STRING section;
   LONG pos, amtentries, i, j;
   UBYTE buffer[80], style[40];

   if (!Args) return(ObjectError(Self, MT_SelectFont, ERR_Args));

   DPrintF("~SelectFont()", "%s", Args->Name);

   if (AccessObject(Self->prvConfigID, 5000, &config) != ERR_Okay) {
      ObjectError(Self, MT_SelectFont, ERR_AccessObject);
      StepBack();
      return(ERR_AccessObject);
   }

   if (GetFields(config, FID_AmtEntries|TLONG, &amtentries,
                         FID_Entries|TPTR,     &entries,
                         TAGEND) != ERR_Okay) {
      ReleaseObject(config);
      ObjectError(Self, MT_SelectFont, ERR_NoData);
      StepBack();
      return(ERR_NoData);
   }

   /*** Find the config section that we should be interested in ***/

   section = NULL;
   for (pos=0; pos < amtentries; pos++) {
      if ((StrCompare(entries[pos].Item, "Name", 0, STR_MATCHLENGTH) IS ERR_Okay) AND
          (StrCompare(entries[pos].Data, Args->Name, 0, STR_MATCHLENGTH) IS ERR_Okay)) {
         section = entries[pos].Section;
         break;
      }
   }

   if (!section) {
      DPrintF("@FontServer:","The font \"%s\" is not installed on this system.", Args->Name);

      /*** Try to default to Helvete ***/

      for (pos=0; pos < amtentries; pos++) {
         if ((StrCompare(entries[pos].Item, "Name", 0, STR_MATCHLENGTH) IS ERR_Okay) AND
             (StrCompare(entries[pos].Data, "Helvete", 0, STR_MATCHLENGTH) IS ERR_Okay)) {
            section = entries[pos].Section;
            break;
         }
      }

      if (!section) {
         ReleaseObject(config);
         StepBack();
         return(ERR_Search);
      }
   }

   if (((Args->Point < 12) OR (Args->Flags & FTF_PREFERFIXED)) AND (!(Args->Flags & FTF_PREFERSCALED))) {
      /*** Try to find a fixed font first ***/

      StrFormat(buffer, sizeof(buffer), "Fixed:%s", Args->Style);
      read.Section = section;
      read.Item    = buffer;
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         Args->Location = StrClone(read.Data);
         ReleaseObject(config);
         StepBack();
         return(ERR_Okay);
      }

      if (StrCompare("Regular", Args->Style, 0, STR_MATCHLENGTH) != ERR_Okay) {
         read.Section = section;
         read.Item    = "Fixed:Regular";
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }
      }

      /*** Try for a scaled font ***/

      if (!(Args->Flags & FTF_PREFERFIXED)) {
         StrFormat(buffer, sizeof(buffer), "Scale:%s", Args->Style);
         read.Section = section;
         read.Item    = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }

         if (StrCompare("Regular", Args->Style, 0, STR_MATCHLENGTH) != ERR_Okay) {
            read.Section = section;
            read.Item    = "Scale:Regular";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               Args->Location = StrClone(read.Data);
               ReleaseObject(config);
               StepBack();
               return(ERR_Okay);
            }
         }
      }

      /*** A regular font style in either format does not exist, so choose the first style that is listed ***/

      read.Section = section;
      read.Item    = "Styles";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         j = 0;
         for (i=0; (read.Data[i]) AND (read.Data[i] != ',') AND (j < sizeof(style)-1); i++) style[j++] = read.Data[i];
         style[j] = 0;

         StrFormat(buffer, sizeof(buffer), "Fixed:%s", style);
         read.Section = section;
         read.Item    = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }

         if (!(Args->Flags & FTF_PREFERFIXED)) {
            StrFormat(buffer, sizeof(buffer), "Scale:%s", style);
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               Args->Location = StrClone(read.Data);
               ReleaseObject(config);
               StepBack();
               return(ERR_Okay);
            }
         }
      }
   }
   else {
      /*** Try to find a scalable font first ***/

      StrFormat(buffer, sizeof(buffer), "Scale:%s", Args->Style);
      read.Section = section;
      read.Item    = buffer;
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         Args->Location = StrClone(read.Data);
         ReleaseObject(config);
         StepBack();
         return(ERR_Okay);
      }

      if (StrCompare("Regular", Args->Style, 0, STR_MATCHLENGTH) != ERR_Okay) {
         read.Section = section;
         read.Item    = "Scale:Regular";
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }
      }

      if (!(Args->Flags & FTF_PREFERSCALED)) {
         StrFormat(buffer, sizeof(buffer), "Fixed:%s", Args->Style);
         read.Section = section;
         read.Item    = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }

         if (StrCompare("Regular", Args->Style, 0, STR_MATCHLENGTH) != ERR_Okay) {
            read.Section = section;
            read.Item    = "Fixed:Regular";
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               Args->Location = StrClone(read.Data);
               ReleaseObject(config);
               StepBack();
               return(ERR_Okay);
            }
         }
      }

      /*** A regular font style in either format does not exist, so choose the first style that is listed ***/

      read.Section = section;
      read.Item    = "Styles";
      if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
         j = 0;
         for (i=0; (read.Data[i]) AND (read.Data[i] != ',') AND (j < sizeof(style)-1); i++) style[j++] = read.Data[i];
         style[j] = 0;

         StrFormat(buffer, sizeof(buffer), "Scale:%s", style);
         read.Section = section;
         read.Item    = buffer;
         if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
            Args->Location = StrClone(read.Data);
            ReleaseObject(config);
            StepBack();
            return(ERR_Okay);
         }

         if (!(Args->Flags & FTF_PREFERSCALED)) {
            StrFormat(buffer, sizeof(buffer), "Fixed:%s", style);
            if (Action(MT_ReadConfig, config, &read) IS ERR_Okay) {
               Args->Location = StrClone(read.Data);
               ReleaseObject(config);
               StepBack();
               return(ERR_Okay);
            }
         }
      }
   }

   ReleaseObject(config);
   StepBack();
   return(ERR_Search);
}

/*****************************************************************************

###FIELD###
Name:   TotalFonts
Short:  Reflects the total number of fonts installed in the system.
Type:   LONG
Status: Get

###DESCRIPTION###

###END###

*****************************************************************************/

static ERROR GET_TotalFonts(struct FontServer *Self, LONG *Value)
{
   OBJECTPTR config;

   if (AccessObject(Self->prvConfigID, 3000, &config) IS ERR_Okay) {
      GetField(config, FID_TotalSections, FT_LONG, Value);
      ReleaseObject(config);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}
