#ifndef FILES_FILE_H
#define FILES_FILE_H TRUE

/*
**   file.h
**
**   (C) Copyright 1996-2003 Rocklyte Systems.
**       All Rights Reserved.
*/

#define VER_FILE  ((FLOAT)1.0)

/*****************************************************************************
** File Flags
*/

#define FL_WRITE       0x00000001
#define FL_FIND        0x00000002
#define FL_NEW         0x00000004
#define FL_READ        0x00000008
#define FL_DIRECTORY   0x00000010
#define FL_APPROXIMATE 0x00000020
#define FL_LINK        0x00000040
#define FL_BUFFER      0x00000080
#define FL_LOOP        0x00000100
#define FL_FILE        0x00000200
#define FL_RESETDATE   0x00000400

#define FL_SYMLINK     FL_LINK

/*****************************************************************************
** Permission Flags
*/

#define PERMIT_READ          0x00000001 /* Owner can read */
#define PERMIT_WRITE         0x00000002 /* Owner can write */
#define PERMIT_EXEC          0x00000004 /* Owner can execute */
#define PERMIT_DELETE        0x00000008 /* Owner can delete */
#define PERMIT_GROUP_READ    0x00000010 /* Group can read */
#define PERMIT_GROUP_WRITE   0x00000020 /* Group can write */
#define PERMIT_GROUP_EXEC    0x00000040 /* Group can execute */
#define PERMIT_GROUP_DELETE  0x00000080 /* Group can delete */
#define PERMIT_OTHERS_READ   0x00000100 /* Others can read */
#define PERMIT_OTHERS_WRITE  0x00000200 /* Others can write */
#define PERMIT_OTHERS_EXEC   0x00000400 /* Others can execute */
#define PERMIT_OTHERS_DELETE 0x00000800 /* Others can delete */
#define PERMIT_HIDDEN        0x00001000 /* Hidden from everybody */
#define PERMIT_ARCHIVE       0x00002000 /* Archive the file for backup */
#define PERMIT_PASSWORD      0x00004000 /* File is password protected */
#define PERMIT_USERID        0x00008000 /* Allows executables to run with a set user id */
#define PERMIT_GROUPID       0x00010000 /* Allows executables to run with a set group id */

#define PERMIT_EVERYONE_READ   (PERMIT_READ | PERMIT_GROUP_READ | PERMIT_OTHERS_READ)
#define PERMIT_EVERYONE_WRITE  (PERMIT_WRITE | PERMIT_GROUP_WRITE | PERMIT_OTHERS_WRITE)
#define PERMIT_EVERYONE_EXEC   (PERMIT_EXEC | PERMIT_GROUP_EXEC | PERMIT_OTHERS_EXEC)
#define PERMIT_EVERYONE_DELETE (PERMIT_DELETE | PERMIT_GROUP_DELETE | PERMIT_OTHERS_DELETE)

#define PERMIT_ALL_READ   PERMIT_EVERYONE_READ   
#define PERMIT_ALL_WRITE  PERMIT_EVERYONE_WRITE  
#define PERMIT_ALL_EXEC   PERMIT_EVERYONE_EXEC   
#define PERMIT_ALL_DELETE PERMIT_EVERYONE_DELETE 

#define PERMIT_EVERYONE_ACCESS (PERMIT_EVERYONE_READ|PERMIT_EVERYONE_WRITE|PERMIT_EVERYONE_EXEC|PERMIT_EVERYONE_DELETE)
#define PERMIT_EVERYONE_READWRITE (PERMIT_EVERYONE_READ | PERMIT_EVERYONE_WRITE)

#define PERMIT_USER      (PERMIT_READ | PERMIT_WRITE | PERMIT_EXEC | PERMIT_DELETE)
#define PERMIT_GROUP     (PERMIT_GROUP_READ | PERMIT_GROUP_WRITE | PERMIT_GROUP_EXEC | PERMIT_GROUP_DELETE)
#define PERMIT_OTHERS    (PERMIT_OTHERS_READ | PERMIT_OTHERS_WRITE | PERMIT_OTHERS_EXEC | PERMIT_OTHERS_DELETE)

/*****************************************************************************
** AnalyseLocation() values.
*/

#define LOC_DIRECTORY 1
#define LOC_ASSIGN    2
#define LOC_FILE      3

/*****************************************************************************
** File Methods.
*/

#define MT_ReadFileWord  -1
#define MT_ReadFileLong  -2
#define MT_DeleteFile    -3
#define MT_MoveFile      -4
#define MT_CopyFile      -5
#define MT_WriteFileWord -6
#define MT_WriteFileLong -7
#define MT_ReadLE        -8
#define MT_ReadBE        -9
#define MT_SetFileDate   -10

struct mtReadLE { LONG Size; LARGE Result; };
struct mtReadBE { LONG Size; LARGE Result; };
struct mtReadFileWord { LONG Result; };
struct mtReadFileLong { LONG Result; };
struct mtWriteFileWord { LONG Value; };
struct mtWriteFileLong { LONG Value; };
struct mtMoveFile { STRING Destination; };
struct mtCopyFile { STRING Destination; };
struct mtSetFileDate { LONG Year; LONG Month; LONG Day; LONG Hour; LONG Minute; LONG Second; };

/*****************************************************************************
** File structure
*/

struct FileTime {
   LONG Year;
   LONG Month;
   LONG Day;
   LONG Hour;
   LONG Minute;
   LONG Second;
};

struct File {
   OBJECT_HEADER         /* Class header */
   LARGE  Position;      /* Current position in file */
   LONG   Flags;         /* File opening flags */
   LONG   Static;        /* */
   BYTE   *Buffer;       /* */
   struct FileTime Time;

   PRIVATE_FIELDS
   LARGE Size;
   LONG  Handle;
   #ifdef _WIN32
      LONG  Stream;
   #else
      APTR  Stream;
   #endif
   STRING   Location;
   MEMORYID LocationMID;
   MEMORYID BufferMID;
   LONG     Stat;
   LONG     Permissions;
};

#endif /* FILES_FILE_H */
